#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
from .index import IndexUpdate
from .metadata import MetadataUpdate
from .apk import ApkUpdate
from .src import SrcUpdate

LOGGER = logging.getLogger('update.Update')
class Update(object):
    ''' handels downloading of repo related data '''
    def __init__(self, config, max_workers=10, head_timeout=10, index_timeout=60, download_timeout=60):
        self.__config = config
        self.__head_timeout = head_timeout
        self.__index_timeout = index_timeout
        self.__download_timeout = download_timeout
        self.__max_workers = max_workers
        self.__index = IndexUpdate(config, head_timeout=head_timeout, index_timeout=index_timeout, max_workers=max_workers)
        self.__meta = None
        self.__apk = ApkUpdate(config, download_timeout=download_timeout, max_workers=max_workers)
        self.__src = SrcUpdate(config, download_timeout=download_timeout, max_workers=max_workers)

    def index(self):
        self.__index.download(*self.__index.required(self.__config.repos, timeout=self.__head_timeout), timeout=self.__index_timeout)
        return self

    def metadata(self):
        self.__meta = MetadataUpdate(self.__config, download_timeout=self.__download_timeout, max_workers=self.__max_workers)
        self.__meta.update_yaml()
        self.__meta.update_assets()
        return self

    def apk(self):
        self.__apk.update()
        return self

    def src(self):
        self.__src.update()
        return self



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
try:
    from collections.abc import MutableMapping
except ImportError:
    from collections import MutableMapping
import copy
import json
import os, os.path
import hashlib
try:
    from urllib.parse import urlparse, urljoin
except ImportError:
    from urlparse import urlparse, urljoin
import shutil
from tempfile import NamedTemporaryFile
from .metadata import Metadata
from .index import Index
from ..json import GenericJSONEncoder


LOGGER = logging.getLogger('model.RepoConfig')
class RepoConfig(MutableMapping):
    EMPTY = {"apps":[]}

    def __init__(self, url, cfg, config):
        self.__config = config
        self.__url = self.__clean_url(url)
        # py3.5 only :( -> self.__store = {**RepoConfig.EMPTY, **cfg}
        self.__store = RepoConfig.EMPTY.copy()
        self.__store.update(cfg)
        self.__store['id'] = hashlib.sha1(str(self.__url).encode('UTF-8')).hexdigest()

    def __clean_url(self, url):
        purl = urlparse(url)
        url = "%s://%s%s"%(purl.scheme, purl.netloc, purl.path)
        if url.endswith('index.xml'):
            url = url[:-9]
        if url.endswith('index.jar'):
            url = url[:-9]
        if url.endswith('index-v1.jar'):
            url = url[:-12]
        if url.endswith('index-v1.json'):
            url = url[:-13]
        if not url.endswith('/'):
            url += '/'
        return url

    @property
    def index(self):
        try:
            return self.__config.index(self.url)
        except KeyError as keyerr:
            LOGGER.error(str(keyerr))
        return None

    @property
    def auth(self):
        auth = self.__store.get('auth', None)
        if auth is None or len(auth) != 2:
            return None
        return (auth[0], auth[1])

    @property
    def verify(self):
        return self.__store.get('ssl_verify', True)

    # pylint: disable=C0103
    @property
    def id(self):
        return self.__store.get('id', None)

    @property
    def key(self):
        return self.url

    @property
    def url(self):
        return str(self.__url)

    @property
    def hash(self):
        return self.__store.get('hash', None)

    @property
    def filename(self):
        return os.path.join(self.__config.cache_dir, self.id+".cache")

    @property
    def format(self):
        if 'format' in self.__store:
            return self.__store['format']
        return None

    @property
    def url_index(self):
        return urljoin(self.__url, 'index.jar')

    @property
    def url_index_v1(self):
        return urljoin(self.__url, 'index-v1.jar')

    @property
    def src_download(self):
        if 'src_download' in self.__store:
            return self.__store['src_download'] is True
        return self.__config.src_download is True

    @property
    def metadata_download(self):
        if 'metadata_download' in self.__store:
            return self.__store['metadata_download'] is True
        return self.__config.metadata_download is True

    @property
    def default_locale(self):
        if 'default_locale' in self.__store:
            return str(self.__store['default_locale'])
        return self.__config.default_locale

    @property
    def apps(self):
        if 'apps' in self.__store:
            if isinstance(self.__store['apps'], str):
                yield self.__store['apps']
            elif isinstance(self.__store['apps'], list):
                for app in self.__store['apps']:
                    yield app

    @property
    def __json__(self):
        ''' make RepoConfig json serializable '''
        return self.__store

    def __repr__(self): return "<RepoConfig: %s>: %s"%(self.__url, self.__store)

    #######################
    # implement hash
    #######################
    def __hash__(self):
        return self.__url.__hash__()

    #######################
    # implement "dict"
    #######################
    def __getitem__(self, key):
        return self.__store[key]
    def __setitem__(self, key, value):
        self.__store[key] = value
    def __delitem__(self, key):
        del self.__store[key]
    def __iter__(self):
        return iter(self.__store)
    def __len__(self):
        return len(self.__store)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
'''
Manages updates
'''

from .update import Update
from .index import IndexUpdate
from .metadata import MetadataUpdate
from .apk import ApkUpdate
from .src import SrcUpdate

__all__ = ['Update', 'IndexUpdate', 'MetadataUpdate', 'ApkUpdate', 'SrcUpdate']



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
from json import JSONEncoder


LOGGER = logging.getLogger('json.GenericJSONEncoder')
class GenericJSONEncoder(JSONEncoder):
    ''' custom encoder so we can serialize Config to json '''
    def default(self, obj):
        if hasattr(obj, '__json__'):
            return getattr(obj, '__json__')
        return JSONEncoder.default(self, obj)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging


LOGGER = logging.getLogger('update.SrcUpdate')
class SrcUpdate(object):
    def __init__(self, config, download_timeout=600, max_workers=10):
        self.__config = config
        self.__download_timeout = download_timeout
        self.__max_workers = max_workers

    def update(self):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

try:
    from collections.abc import MutableMapping
except ImportError:
    from collections import MutableMapping
import re
import json
import xml.etree.ElementTree as ET
from time import mktime, strptime
import logging
try:
    from urllib.parse import urljoin
except ImportError:
    from urlparse import urljoin
from tempfile import NamedTemporaryFile
import shutil
from ..json import GenericJSONEncoder


LOGGER = logging.getLogger('model.Index')
class Index(MutableMapping):
    """
    Main Class responsible for handling the f-droid's index.xml and
    index-v1.json files.
    """

    def __init__(self, key=None, filename=None, format='json', default_locale='en-US', store=dict()):
        self.__key = key
        self.__filename = filename
        self.__format = format
        self.__default_locale = default_locale
        self.__store = store

    @classmethod
    def from_json(cls, source, **kwargs):
        if not hasattr(source, "read"):
            source = open(source, "r")
        kwargs['format'] = 'json'
        kwargs['filename'] = source.name
        return cls(**kwargs).load(source)

    @classmethod
    def from_xml(cls, source, **kwargs):
        if not hasattr(source, "read"):
            source = open(source, "r")
        kwargs['format'] = 'xml'
        kwargs['filename'] = source.name
        return cls(**kwargs).load(source)

    @property
    def filename(self):
        return str(self.__filename)

    @property
    def key(self):
        return str(self.__key)

    @property
    def repo_url(self):
        return str(self.__key)

    @property
    def format(self):
        return str(self.__format)

    @property
    def default_locale(self):
        return str(self.__default_locale)

    def find_appids(self, key):
        if key is None:
            raise KeyError("key must not be empty")
        ret_val = set()
        if key.startswith('regex:'):
            regexc = re.compile(key[6:], re.I | re.S)
            for app in self.__store['apps']:
                match = regexc.match(app['packageName'])
                if not match is None:
                    ret_val.add(app['packageName'])
        elif key in ['*', '.*', 'all']:
            for app in self.__store['apps']:
                ret_val.add(app['packageName'])
        else:
            for app in self.__store['apps']:
                if app['packageName'] == key:
                    ret_val.add(app['packageName'])
        return list(ret_val)

    def defaultnum(self, value):
        if value.isnumeric():
            return int(value)
        return value

    def custom_func(self, xpath):
        if not xpath is None:
            regexc = re.compile(r'(.*?):(.*)$', re.I | re.S)
            match = regexc.match(xpath)
            if not match is None:
                return (match.group(1), match.group(2))
            return (xpath, None)
        return (None, None)

    def run_custom_func(self, method, value):
        if method is None or value is None:
            return None
        try:
            return eval(method, {'mktime': mktime, 'strptime': strptime}, {'value': value})
        except:
            LOGGER.exception("error in eval(%s)", method)

    def xmlattr(self, node, xml_node, key=None, xpath=None):
        if xpath is None and key is None:
            return None
        if xpath is None and not key is None:
            xpath = key
        value = self.xmlattrval(xml_node, xpath)
        if not value is None:
            if not key is None:
                node[key] = value
            else:
                node = value

    def xmlattrval(self, xml_node, xpath):
        if not xml_node is None:
            regexc = re.compile(r'^.*?\[@([^\]]+)\]$')
            match = regexc.match(xpath)
            if not match is None:
                attr_name = m.group(1)
            else:
                attr_name = xpath
            (xpath, func) = self.custom_func(xpath)
            if not xpath is None:
                xml_node = xml_node.find(xpath)
            if not xml_node is None:
                value = xml_node.get(attr_name, None)
                #print(xml_node,attr_name,value)
                if not value is None:
                    if not func is None:
                        return self.run_custom_func(func, self.defaultnum(value.strip()))
                    else:
                        return self.defaultnum(value.strip())
        return None

    def xmltext(self, node, xml_node, key=None, xpath=None):
        if xpath is None and key is None:
            return None
        if xpath is None and not key is None:
            xpath = key
        value = self.xmltextval(xml_node, xpath)
        if not value is None:
            if not key is None:
                node[key] = value
            else:
                node = value

    def xmltextval(self, xml_node, xpath=None):
        (xpath, func) = self.custom_func(xpath)
        if not xpath is None:
            xml_node = xml_node.find(xpath)
        if not xml_node is None:
            value = xml_node.text
            if not value is None:
                if not func is None:
                    return self.run_custom_func(func, self.defaultnum(value.strip()))
                else:
                    return self.defaultnum(value.strip())

    def xmllist(self, root, xpath):
        ret_val = []
        regexc = re.compile(r'^.*?\[@([^\]]+)\]$')
        match = regexc.match(xpath)
        if not match is None:
            attr_name = match.group(1)
            matches = root.findall(xpath)
            if not matches is None:
                for match in matches:
                    value = match.get(attr_name, None)
                    if not value is None:
                        ret_val.append(self.defaultnum(value.strip()))
        else:
            matches = root.findall(xpath)
            if not matches is None:
                for match in matches:
                    value = self.xmltextval(match)
                    if not value is None:
                        ret_val.append(value)
        return ret_val

    def convert(self, root):
        self.__store = {}
        store = self.__store
        repo = root.find('repo')
        if not repo is None:
            srepo = store['repo'] = {}
            self.xmlattr(srepo, root, 'timestamp', './repo[@timestamp]')
            self.xmlattr(srepo, root, 'version', './repo[@version]')
            self.xmlattr(srepo, root, 'maxage', './repo[@maxage]')
            self.xmlattr(srepo, root, 'name', './repo[@name]')
            self.xmlattr(srepo, root, 'icon', './repo[@icon]')
            self.xmlattr(srepo, root, 'url', './repo[@address]')
            self.xmltext(srepo, root, 'description', './description')
            srepo['mirrors'] = self.xmllist(repo, './mirror')
        requ = store['requests'] = {}
        requ['install'] = self.xmllist(root, './install[@packageName]')
        requ['uninstall'] = self.xmllist(root, './uninstall[@packageName]')
        apps = store['apps'] = []
        pkgs = store['packages'] = {}
        applications = root.findall('application[@id]')
        for xmlapp in applications:
            app_val = {}
            self.xmltext(app_val, xmlapp, 'authorEmail', './email')
            self.xmltext(app_val, xmlapp, 'authorName', './author')
            self.xmltext(app_val, xmlapp, 'authorWebSite', './web')
            self.xmltext(app_val, xmlapp, 'bitcoin', './bitcoin')
            self.xmltext(app_val, xmlapp, 'donate', './donate')
            self.xmltext(app_val, xmlapp, 'flattr', './flattr')
            self.xmltext(app_val, xmlapp, 'liberapay', './liberapay')
            self.xmltext(app_val, xmlapp, 'litecoin', './litecoin')
            cat = self.xmltextval(xmlapp, 'categories')
            if not cat is None and len(cat) > 0:
                app_val['categories'] = cat.split(',')
            afaet = self.xmltextval(xmlapp, 'antiFeatures')
            if not afaet is None and len(afaet) > 0:
                app_val['antiFeatures'] = afaet.split(',')
            self.xmltext(app_val, xmlapp, 'suggestedVersionName',
                         './marketversion')
            self.xmltext(app_val, xmlapp, 'suggestedVersionCode',
                         './marketvercode')
            self.xmltext(app_val, xmlapp, 'issueTracker', './tracker')
            self.xmltext(app_val, xmlapp, 'changelog', './changelog')
            self.xmltext(app_val, xmlapp, 'license', './license')
            self.xmltext(app_val, xmlapp, 'name', './name')
            self.xmltext(app_val, xmlapp, 'sourceCode', './source')
            self.xmltext(app_val, xmlapp, 'webSite', './web')
            self.xmltext(app_val, xmlapp, 'added',
                         './added:int(mktime(strptime(value, "%Y-%m-%d")))')
            self.xmltext(app_val, xmlapp, 'icon', './icon')
            self.xmltext(app_val, xmlapp, 'packageName', './id')
            self.xmltext(app_val, xmlapp, 'lastUpdated',
                         './lastupdated:int(mktime(strptime(value, "%Y-%m-%d")))')
            loc = app_val['localized'] = {}
            dloc = loc[self.default_locale] = {}
            self.xmltext(dloc, xmlapp, 'description', './desc')
            self.xmltext(dloc, xmlapp, 'summary', './summary')

            app_id = self.xmltextval(xmlapp, './id')
            if not app_id is None:
                pkg = pkgs[app_id] = []
                packages = xmlapp.findall('./package')
                for xmlpkg in packages:
                    pkg_val = {}
                    self.xmltext(
                        pkg_val, xmlpkg, 'added', './added:int(mktime(strptime(value, "%Y-%m-%d")))')
                    self.xmltext(pkg_val, xmlpkg, 'apkName', './apkname')
                    self.xmltext(pkg_val, xmlpkg, 'hash', './hash')
                    self.xmlattr(pkg_val, xmlpkg, 'hashType', './hash[@type]')
                    self.xmltext(pkg_val, xmlpkg, 'minSdkVersion', './sdkver')
                    self.xmltext(pkg_val, xmlpkg, 'targetSdkVersion',
                                 './targetSdkVersion')
                    pkg_val['packageName'] = app_id
                    self.xmltext(pkg_val, xmlpkg, 'sig', './sig')
                    self.xmltext(pkg_val, xmlpkg, 'versionName', './version')
                    self.xmltext(pkg_val, xmlpkg, 'versionCode',
                                 './versioncode')
                    self.xmltext(pkg_val, xmlpkg, 'size', './size')
                    perm = xmlpkg.find('permissions')
                    if not perm is None:
                        user_permission = pkg_val['uses-permission'] = []
                        user_permission_split = perm.text.split(",")
                        for user_permission_str in user_permission_split:
                            user_permission.append(['android.permission.'+user_permission_str.strip(), None])
                    if len(pkg_val) > 0:
                        pkg.append(pkg_val)
            if len(app_val) > 1:
                apps.append(app_val)

    def monkeypatch(self):
        ''' fixup metadata paths -> url '''
        if not '_monkeypatched' in self.__store:
            if self.__key is None:
                raise AttributeError('key not defined')
            self.__store['_monkeypatched'] = True
            if 'packages' in self.__store:
                for key in self.__store['packages']:
                    for pkg in self.__store['packages'][key]:
                        pkg['apkName'] = urljoin(self.__key, pkg['apkName'])
                        if not pkg.get('srcname') is None:
                            pkg['srcname'] = urljoin(
                                self.__key, pkg['srcname'])

            if 'apps' in self.__store:
                for app in self.__store['apps']:
                    if 'icon' in app:
                        app['icon'] = urljoin(self.__key, "icons/"+app['icon'])
                    locs = app.get('localized', {})
                    for k in locs.keys():
                        loc = locs[k]
                        if 'icon' in loc:
                            loc['icon'] = urljoin(
                                self.__key, app['packageName']+'/'+k+'/'+loc['icon'])
                        if 'featureGraphic' in loc:
                            loc['featureGraphic'] = urljoin(
                                self.__key, app['packageName']+'/'+k+'/'+loc['featureGraphic'])
                        if 'promoGraphic' in loc:
                            loc['promoGraphic'] = urljoin(
                                self.__key, app['packageName']+'/'+k+'/'+loc['promoGraphic'])
                        if 'tvBanner' in loc:
                            loc['tvBanner'] = urljoin(
                                self.__key, app['packageName']+'/'+k+'/'+loc['tvBanner'])
                        if 'phoneScreenshots' in loc:
                            loc['phoneScreenshots'] = [urljoin(
                                self.__key, app['packageName']+'/'+k+'/phoneScreenshots/'+value) for value in loc['phoneScreenshots']]
                        if 'sevenInchScreenshots' in loc:
                            loc['sevenInchScreenshots'] = [urljoin(
                                self.__key, app['packageName']+'/'+k+'/sevenInchScreenshots/'+value) for value in loc['sevenInchScreenshots']]
                        if 'tenInchScreenshots' in loc:
                            loc['tenInchScreenshots'] = [urljoin(
                                self.__key, app['packageName']+'/'+k+'/tenInchScreenshots/'+value) for value in loc['tenInchScreenshots']]
                        if 'tvScreenshots' in loc:
                            loc['tvScreenshots'] = [urljoin(
                                self.__key, app['packageName']+'/'+k+'/tvScreenshots/'+value) for value in loc['tvScreenshots']]
                        if 'wearScreenshots' in loc:
                            loc['wearScreenshots'] = [urljoin(
                                self.__key, app['packageName']+'/'+k+'/wearScreenshots/'+value) for value in loc['wearScreenshots']]
        return self

    def load(self, file, format=None):
        if not hasattr(file, 'read'):
            file = open(file, 'r')
        if not format is None:
            self.__format = format

        if self.__format == 'json':
            with file as idxfl:
                self.__store = json.load(idxfl)
        elif self.__format == 'xml':
            # we have no json verison we need som transformation to apply
            tree = ET.parse(file)
            file.close()
            root = tree.getroot()
            self.convert(root)
            self.__format = 'json'
        else:
            # try loading json anyway
            try:
                self.load(file, format='json')
            except:
                self.load(file, format='xml')
        return self

    def save(self, filename=None):
        if not filename is None:
            self.__filename = filename
        with NamedTemporaryFile(mode='w') as tmp:
            json.dump(self.__store, tmp, sort_keys=True,
                      indent=4, cls=GenericJSONEncoder)
            tmp.flush()
            shutil.copy(tmp.name, self.__filename)
        return self

    def __repr__(self):
        return "<Index: %s>" % str(json.dumps(self.__store, indent=4))

    @property
    def __json__(self):
        ''' make Index json serializable '''
        return self.__store
    #######################
    # implement "dict"
    #######################

    def __getitem__(self, key):
        return self.__store[key]

    def __setitem__(self, key, value):
        self.__store[key] = value

    def __delitem__(self, key):
        del self.__store[key]

    def __iter__(self):
        return iter(self.__store)

    def __len__(self):
        return len(self.__store)
    #######################
    # implement "with"
    #######################

    def __enter__(self):
        self.load(self.__filename)
        return self

    def __exit__(self, type, value, traceback):
        self.save()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
import requests


LOGGER = logging.getLogger('update.Selector')
class Selector(object):
    def __init__(self, config):
        self.__config = config

    def __meta_repos(self):
        for repo in self.__config.repos:
            if not 'error' in repo: # ignore repos with download error
                yield repo

    def all_apps(self, dupes=False, session=None):
        yielded = set()
        for repo in self.__meta_repos():
            Selector.apply_session_settings(repo, session)
            for selector in repo.apps:
                if not repo.index is None:
                    for appid in repo.index.find_appids(selector):
                        if not appid in yielded or dupes:
                            yielded.add(appid)
                            yield (repo, appid)

    @staticmethod
    def apply_session_settings(repo, session):
        ''' apply security settings for basic auth and ssl verification '''
        if not session is None:
            if not repo.auth is None or repo.verify is False:
                thread_session = requests.Session()
                thread_session.auth = repo.auth
                thread_session.verify = repo.verify
                session.map(repo.url, thread_session)



import os
import unittest
from fdroid_dl.model import Config
try:
    from unittest.mock import patch
except ImportError:
    from mock import patch

class ConfigTestSuite(unittest.TestCase):

    @patch('os.makedirs')
    @patch('os.path.exists')
    @patch.object(Config, 'save')
    def test_prepare_dirs(self, save, exists, makedirs):
        exists.return_value = False
        with Config() as c:
            exists.assert_any_call('./repo')
            exists.assert_any_call('./metadata')
            exists.assert_any_call('.cache')
            makedirs.assert_any_call('./repo')
            makedirs.assert_any_call('./metadata')
            makedirs.assert_any_call('.cache')
            self.assertEqual(c.filename, 'fdroid-dl.json')
            self.assertEqual(c.repo_dir, './repo')
            self.assertEqual(c.metadata_dir, './metadata')
            self.assertEqual(c.cache_dir, '.cache')
            self.assertIsNot(c.get('f-droid'), None)
            self.assertEqual([x for x in iter(c)],
                             [x for x in iter(Config.DEFAULTS)])
            assert not save.called
        assert save.called



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
import time
import os.path
from datetime import timedelta
try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse
from .selector import Selector
from ..download import FuturesSessionVerifiedDownload, FuturesSessionFlex

LOGGER = logging.getLogger('update.ApkUpdate')
class ApkUpdate(Selector):
    def __init__(self, config, download_timeout=600, max_workers=10):
        super(ApkUpdate, self).__init__(config)
        self.__config = config
        self.__download_timeout = download_timeout
        self.__max_workers = max_workers

    def all_packages(self, session=None):
        downloads = {}
        logging.info("collecting apps to download")
        start = time.time()
        apkcnt = 0
        # search pkgs
        for repo, appid in self.all_apps(dupes=True):
            Selector.apply_session_settings(repo, session)
            packages = repo.index.get('packages', {}).get(appid, [])
            if not appid in downloads:
                downloads[appid] = []
            for pkg in packages:
                if not pkg.get('apkName') is None and not pkg.get('hash') is None and not pkg.get('hashType') is None:
                    downloads[appid].append(pkg)
                    apkcnt += 1
        elapsed = time.time() - start
        logging.info("found (%s) apk files to download (%s)", apkcnt, timedelta(seconds=elapsed))

        for appid in downloads.keys():
            # sort newest first
            packages = sorted(downloads[appid], key=lambda e: e.get('versionCode', 0), reverse=True)
            for idx, pkg in enumerate(packages):
                if idx >= self.__config.apk_versions:
                    break # only download number of confgured files
                url = pkg.get('apkName')
                filename = os.path.basename(str(urlparse(url).path))
                filepath = os.path.join(self.__config.repo_dir, filename)
                yield (url, filepath, pkg.get('hash'), pkg.get('hashType'))

    def update(self):
        meta = self.__config.metadata
        LOGGER.info("UPDATING apk files")
        start = time.time()
        cnt = 0
        ecnt = 0
        with FuturesSessionVerifiedDownload(max_workers=self.__max_workers) as session:
            for url, filename, fhash, hash_type in self.all_packages(session=session):
                if os.path.exists(filename):
                    start2 = time.time()
                    if FuturesSessionVerifiedDownload.verify(filename, hash_type, fhash):
                        elapsed = time.time() - start2
                        LOGGER.info("hash verified %s [%s] (%s) ✔", os.path.basename(filename), timedelta(seconds=elapsed), FuturesSessionFlex.h_size(os.stat(filename).st_size))
                    else:
                        session.download(url, filename, timeout=self.__download_timeout, hash=fhash, hash_type=hash_type)
                else:
                    session.download(url, filename, timeout=self.__download_timeout, hash=fhash, hash_type=hash_type)

            dlsum = 0
            for success, filename, dbytes, hbytes, elapsed in session.completed():
                if success:
                    cnt += 1
                    dlsum += dbytes
                else:
                    ecnt += 1
        elapsed = time.time() - start
        LOGGER.info("UPDATED apk files, files(%s) errors(%s) [%s] (%s)", cnt, ecnt, FuturesSessionFlex.h_size(dlsum), timedelta(seconds=elapsed))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from .config import Config
from .repoconfig import RepoConfig
from .appmetadata import AppMetadata
from .metadata import Metadata
from .index import Index

__all__ = ['Config', 'RepoConfig', 'AppMetadata', 'Metadata', 'Index']



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

try:
    from collections.abc import MutableMapping
except ImportError:
    from collections import MutableMapping
import logging
import json
import hashlib
import copy
import re
from ..json import GenericJSONEncoder
from .index import Index
from .appmetadata import AppMetadata

LOGGER = logging.getLogger('model.Metadata')
class Metadata(MutableMapping):
    def __init__(self, repoman, config={}, default_locale='en-US'):
        self.__repoman = repoman
        self.__config = config
        self.__store = {}
        self.__default_locale = default_locale
        self.__load()

    def __load(self):
        # load from config file
        for key in self.__config.keys():
            self.__config[key] = AppMetadata(key, self.__config[key], default_locale=self.__default_locale)
            self.__store[key] = AppMetadata(key, self.__config[key], default_locale=self.__default_locale)
        # load from index files

    def load(self, index):
        if not isinstance(index, Index) and not isinstance(index, dict) and not isinstance(index, str):
            raise NotImplementedError("I can only load instances of Index at the moment")
        if isinstance(index, str):
            index = self.__repoman.index(index)
        apps = index.get('apps', [])
        for app in apps:
            self.add(app)

    def load_all(self):
        for index in self.__repoman.indices:
            LOGGER.info("loading index: %s", index.key)
            apps = index.get('apps', [])
            for app in apps:
                appid = app.get('packageName', None)
                if not appid is None:
                    self.add(AppMetadata(appid, app, default_locale=self.__default_locale))
            LOGGER.info("loaded index: %s apps: %s", index.key, len(apps))
        return self

    def add(self, appmetadata):
        if appmetadata is None:
            return
        if appmetadata.id is None:
            KeyError('cant add metadata without id! %s'%appmetadata)
        if appmetadata.id in self.__store:
            self.__store[appmetadata.id].merge(appmetadata)
        else:
            self.__store[appmetadata.id] = appmetadata
        return self

    def find_all(self, key):
        if key is None:
            raise KeyError("key must not be empty")
        ret_val = set()
        if key.startswith('regex:'):
            regc = re.compile(key[6:], re.I|re.S)
            for k in self.__store:
                match = regc.match(k)
                if not match is None:
                    ret_val.add(self.__store[k])
        elif key in ['*', '.*', 'all']:
            for k in self.__store:
                ret_val.add(self.__store[k])
        elif key in self:
            ret_val.add(self[key])
        return list(ret_val)

    def __repr__(self):
        return "<Metadata: %s>"%str(json.dumps(self.__store, indent=4, cls=GenericJSONEncoder))

    @property
    def __json__(self):
        ''' make Metadata json serializable '''
        return self.__config

    #######################
    # implement "dict"
    #######################
    def __getitem__(self, key):
        return self.__store[key]
    def __setitem__(self, key, value):
        self.__store[key] = value
    def __delitem__(self, key):
        del self.__store[key]
    def __iter__(self):
        return iter(self.__store)
    def __len__(self):
        return len(self.__store)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
import time
from datetime import timedelta
import os.path
try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse
import yaml
from .selector import Selector
from ..download import FuturesSessionVerifiedDownload


LOGGER = logging.getLogger('update.MetadataUpdate')
class MetadataUpdate(Selector):
    def __init__(self, config, download_timeout=600, max_workers=10):
        super(MetadataUpdate, self).__init__(config)
        self.__config = config
        self.__download_timeout = download_timeout
        self.__max_workers = max_workers
        self.__loaded = False
        self.__load_all()

    def __load_all(self):
        if not self.__is_loaded():
            self.__config.metadata.load_all()
            self.__loaded = True

    def __is_loaded(self):
        return self.__loaded is True

    @staticmethod
    def _setyamlattr(yaml_key, json_key, yaml_data, app_meta):
        value = app_meta.get(json_key)
        if not value is None:
            yaml_data[yaml_key] = value

    def update_yaml(self):
        meta = self.__config.metadata
        LOGGER.info("UPDATING YAML metadata")
        start = time.time()
        cnt = 0
        for repo, appid in self.all_apps():
            try:
                app_meta = meta[appid]
                yaml_file = os.path.join(self.__config.metadata_dir, appid+'.yml')
                yaml_data = {}
                if os.path.exists(yaml_file):
                    with open(yaml_file, 'r') as yfl:
                        yaml_data = yaml.load(yfl, Loader=yaml.SafeLoader)
                    if yaml_data is None:
                        yaml_data = {}
                MetadataUpdate._setyamlattr('Categories', 'categories', yaml_data, app_meta)
                MetadataUpdate._setyamlattr('AuthorName', 'authorName', yaml_data, app_meta)
                MetadataUpdate._setyamlattr('AuthorEmail', 'authorEmail', yaml_data, app_meta)
                MetadataUpdate._setyamlattr('License', 'license', yaml_data, app_meta)
                #self._setyamlattr('Name','name',yaml_data,app_meta)
                MetadataUpdate._setyamlattr('WebSite', 'webSite', yaml_data, app_meta)
                MetadataUpdate._setyamlattr('SourceCode', 'sourceCode', yaml_data, app_meta)
                MetadataUpdate._setyamlattr('IssueTracker', 'issueTracker', yaml_data, app_meta)
                MetadataUpdate._setyamlattr('Changelog', 'changelog', yaml_data, app_meta)
                MetadataUpdate._setyamlattr('Donate', 'donate', yaml_data, app_meta)
                MetadataUpdate._setyamlattr('FlattrID', 'flattr', yaml_data, app_meta)
                MetadataUpdate._setyamlattr('LiberapayID', 'liberapay', yaml_data, app_meta)
                MetadataUpdate._setyamlattr('Bitcoin', 'bitcoin', yaml_data, app_meta)
                MetadataUpdate._setyamlattr('Litecoin', 'litecoin', yaml_data, app_meta)
                MetadataUpdate._setyamlattr('AntiFeatures', 'antiFeatures', yaml_data, app_meta)
                with open(yaml_file, 'w') as stream:
                    yaml.safe_dump(yaml_data, stream, default_flow_style=False, encoding='utf-8', allow_unicode=True)
                cnt += 1
            except Exception as ex:
                if logging.getLogger().isEnabledFor(logging.DEBUG):
                    LOGGER.exception()
                else:
                    LOGGER.warning(str(ex))
        elapsed = time.time() - start
        LOGGER.info("UPDATED YAML metadata, %s files (%s)", cnt, timedelta(seconds=elapsed))

    @staticmethod
    def __write_text(text, filename, foldername):
        if not text is None:
            text = text.encode('utf-8')
            if not os.path.exists(foldername):
                os.makedirs(foldername)
            with open(os.path.join(foldername, filename), "w") as text_file:
                text_file.write(text)

    def __download_image(self, url, filename, session):
        if not url is None:
            session.download(url, filename, timeout=self.__download_timeout)

    def __download_images(self, urls, foldername, session):
        for url in urls:
            filename = os.path.basename(urlparse(url).path)
            filename = os.path.join(foldername, filename)
            self.__download_image(url, filename, session)

    def update_assets(self):
        meta = self.__config.metadata
        LOGGER.info("UPDATING Assets metadata")
        start = time.time()
        cnt = 0
        ecnt = 0
        with FuturesSessionVerifiedDownload(max_workers=self.__max_workers) as session:
            for repo, appid in self.all_apps(session=session):
                try:
                    loc_appid = os.path.join(self.__config.metadata_dir, appid)
                    app_meta = meta[appid]
                    for locale in app_meta.locales:
                        loc_path = os.path.join(loc_appid, locale)
                        imags_path = os.path.join(loc_path, 'images')
                        # TODO: chech if we need to download really all images & clean folder before/after?
                        MetadataUpdate.__write_text(app_meta.full_description(locale), 'full_description.txt', loc_path)
                        MetadataUpdate.__write_text(app_meta.short_description(locale), 'short_description.txt', loc_path)
                        MetadataUpdate.__write_text(app_meta.title(locale), 'title.txt', loc_path)

                        self.__download_image(app_meta.icon(locale), os.path.join(imags_path, 'icon.png'), session)
                        self.__download_image(app_meta.feature_graphic(locale), os.path.join(imags_path, 'featureGraphic.png'), session)
                        self.__download_image(app_meta.promo_graphic(locale), os.path.join(imags_path, 'promoGraphic.png'), session)
                        self.__download_image(app_meta.tv_banner(locale), os.path.join(imags_path, 'tvBanner.png'), session)

                        self.__download_images(app_meta.phone_screenshots(locale), os.path.join(loc_path, 'phoneScreenshots'), session)
                        self.__download_images(app_meta.seven_inch_screenshots(locale), os.path.join(loc_path, 'sevenInchScreenshots'), session)
                        self.__download_images(app_meta.ten_inch_screenshots(locale), os.path.join(loc_path, 'tenInchScreenshots'), session)
                        self.__download_images(app_meta.tv_screenshots(locale), os.path.join(loc_path, 'tvScreenshots'), session)
                        self.__download_images(app_meta.wear_screenshots(locale), os.path.join(loc_path, 'wearScreenshots'), session)
                except Exception:
                    LOGGER.exception("Error processing Asset download for %s", appid)
            for success, filename, bts, hbts, elapsed in session.completed():
                if success:
                    cnt += 1
                else:
                    ecnt += 1
        elapsed = time.time() - start
        LOGGER.info("UPDATED Assets metadata, %s files, %s errors (%s)", cnt, ecnt, timedelta(seconds=elapsed))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
import os.path
from concurrent.futures import as_completed
import requests
from ..download import FuturesSessionFlex
from ..processor import IndexFileProcessor


LOGGER = logging.getLogger('update.IndexUpdate')
class IndexUpdate(object):
    def __init__(self, config, head_timeout=10, index_timeout=60, max_workers=10):
        self.config = config
        self.head_timeout = head_timeout
        self.index_timeout = index_timeout
        self.max_workers = max_workers

    def required(self, repos, timeout=60):
        # try new index
        head_futures = self.__future(repos=repos, timeout=timeout)
        # process result
        (new_index, notfound) = self.__head_response(head_futures)
        # retry old index
        head_futures = self.__future(repos=notfound, attr='url_index', timeout=timeout)
        # process result
        (old_index, notfound) = self.__head_response(head_futures)
        # new_index[] needs to be fetched with index-v1.jar
        # old_index[] needs to be fetched with index.jar
        return (new_index, old_index)

    def download(self, new_index, old_index, timeout=60):
        if new_index is None:
            raise AttributeError('new_index missing')
        if old_index is None:
            raise AttributeError('old_index missing')
        new_futures = self.__future(repos=new_index, timeout=timeout, http_method=FuturesSessionFlex.get, hooks={'response':[FuturesSessionFlex.extract_jar]}, stream=True)
        old_futures = self.__future(repos=old_index, timeout=timeout, http_method=FuturesSessionFlex.get, hooks={'response':[FuturesSessionFlex.extract_jar]}, attr='url_index', stream=True)
        self.__download_response(new_futures)
        self.__download_response(old_futures)

    def __future(self, repos=None, attr='url_index_v1', http_method=FuturesSessionFlex.head, hooks=None, timeout=60, **kwargs):
        if repos is None:
            raise AttributeError('repos missing %s'%repos)
        if hooks is None:
            hooks = {'response': [FuturesSessionFlex.add_hash]}
        futures = []
        with FuturesSessionFlex(max_workers=self.max_workers) as session:
            for repo in repos:
                if not repo.auth is None or repo.verify is False:
                    thread_session = requests.Session()
                    thread_session.auth = repo.auth
                    thread_session.verify = repo.verify
                    session.map(getattr(repo, attr), thread_session)
                request = http_method(session, getattr(repo, attr), hooks=hooks, timeout=timeout, **kwargs)
                request.repo = repo # pass repo ref to future processing
                futures.append(request)
        return futures

    def __as_completed(self, futures):
        for future in as_completed(futures):
            repo = future.repo
            try:
                response = future.result()
                response.raise_for_status()
                yield (repo, response, True)
            except requests.exceptions.HTTPError as ex:
                if ex.response.status_code == 404: # try old index file
                    if logging.getLogger().isEnabledFor(logging.DEBUG):
                        LOGGER.exception()
                    else:
                        LOGGER.warning(str(ex))
                    yield (repo, response, False)
                else:
                    repo['error'] = {
                        'code': ex.response.status_code if isinstance(ex, requests.exceptions.HTTPError) else 600,
                        'msg':str(ex)
                    }
                    LOGGER.exception()

    def __head_response(self, futures):
        success = []
        notfound = []
        for repo, response, okheader in self.__as_completed(futures):
            if okheader:
                # check cache
                LOGGER.info("HEAD %s (%s) ", response.url, response.elapsed)
                if not os.path.exists(repo.filename) or repo.hash != response.hash:
                    if 'error' in repo:
                        del repo['error']
                    repo['hash'] = response.hash
                    success.append(repo)
                    if not os.path.exists(repo.filename):
                        LOGGER.warning("CACHE - (miss) - %s.cache file not found!", repo.id)
                    else:
                        LOGGER.info("CACHE - (miss) - %s - %s)", repo.key, response.hash)
                else:
                    # skip do nothing for cache hits
                    LOGGER.info("CACHE - (hit) - %s - %s)", repo.key, response.hash)
            else:
                notfound.append(repo)
        return (success, notfound)

    def __download_response(self, futures):
        with IndexFileProcessor(max_workers=self.max_workers) as ifp:
            for repo, response, success in self.__as_completed(futures):
                if response.ok:
                    LOGGER.info("DOWNLOADED %s [%s] (%s) ", response.url, response.elapsed, response.h_size)
                    ifp.process(response.index, repo, repo.url, response.h_size)
            for future in ifp.completed():
                (index, elapsed, url, h_size) = future.result()
                repo_name = index.get('repo', {}).get('name')
                LOGGER.info("UPDATED %s - %s [%s] (%s) ", repo_name, url, elapsed, h_size)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import logging
import hashlib
import time
import copy
from datetime import timedelta
from zipfile import ZipFile
from tempfile import NamedTemporaryFile
from requests.adapters import HTTPAdapter
from requests_futures.sessions import FuturesSession


LOGGER = logging.getLogger('download.FuturesSessionFlex')
class FuturesSessionFlex(FuturesSession):
    BLOCKSIZE = 65536
    SUFFIXES = ['B', 'KB', 'MB', 'GB', 'TB', 'PB']

    def __init__(self, max_workers=1, user_agent='Mozilla/5.0 (compatible; Googlebot/2.1; +http://www.google.com/bot.html)', *args, **kwargs):
        kwargs.update({'max_workers': max_workers})
        super(FuturesSessionFlex, self).__init__(*args, **kwargs)
        self.__sessions = {}
        self.__sessions_keys = []

        self.__fs_kwargs = {}
        self.__fs_kwargs.update(kwargs)

        _adapter_kwargs = {'pool_connections': max_workers, 'pool_maxsize': max_workers, 'pool_block': True}
        self.mount('https://', HTTPAdapter(**_adapter_kwargs))
        self.mount('http://', HTTPAdapter(**_adapter_kwargs))
        if self.headers is None:
            self.headers = {}
        self.headers.update({'User-Agent': user_agent})

    def map(self, pattern='http://', session=None):
        ''' if called with session None -> default session for ctor is used '''
        kwargs = copy.deepcopy(self.__fs_kwargs)
        kwargs['session'] = session
        if pattern not in self.__sessions:
            self.__sessions_keys.append(pattern)
        self.__sessions[pattern] = FuturesSessionFlex(*(), **kwargs)
        self.__sessions_keys = sorted(self.__sessions_keys, key=len, reverse=True)

    def set_headers(self, headers):
        self.headers.update(headers)

    @staticmethod
    def h_size(nbytes):
        i = 0
        while nbytes >= 1024 and i < len(FuturesSessionFlex.SUFFIXES)-1:
            nbytes /= 1024.
            i += 1
        fsize = ('%.2f' % nbytes).rstrip('0').rstrip('.')
        return '%s %s' % (fsize, FuturesSessionFlex.SUFFIXES[i])

    # pylint: disable=W0613
    @staticmethod
    def add_size(response, *args, **kwargs):
        if 'Content-Length' in response.headers:
            response.size = int(response.headers.get('Content-Length', 0))
        else:
            LOGGER.warning("Content-Length Header not provided by %s", response.url)
            response.size = len(response.content)
        response.h_size = FuturesSessionFlex.h_size(response.size)
        return response

    # pylint: disable=W0613
    @staticmethod
    def add_hash(response, *args, **kwargs):
        response.hash = None
        if response.ok:
            timestamp = str(time.time())
            cache_var = response.headers.get('Last-Modified', timestamp) + response.headers.get('ETag', timestamp)
            response.hash = hashlib.sha1(str(cache_var).encode('UTF-8')).hexdigest()
        return response

    @staticmethod
    def extract_jar(response, *args, **kwargs):
        if response.ok:
            start = time.time()
            response = FuturesSessionFlex.add_size(response, *args, **kwargs)
            response.index = NamedTemporaryFile()
            with NamedTemporaryFile() as temp_file:
                for chunk in response.iter_content(chunk_size=FuturesSessionFlex.BLOCKSIZE):
                    if chunk:
                        temp_file.write(chunk)
                zip_file = ZipFile(temp_file)
                idxfile = 'index-v1.json' if 'index-v1.json' in zip_file.namelist() else 'index.xml'
                with zip_file.open(idxfile) as file:
                    while True:
                        byte = file.read(FuturesSessionFlex.BLOCKSIZE)
                        if not byte:
                            break
                        response.index.write(byte)
            LOGGER.debug("%s - %s - (%s)", response.index.name, response.url, FuturesSessionFlex.h_size(os.stat(response.index.name).st_size))
            elapsed = time.time() - start
            response.elapsed += timedelta(seconds=elapsed)
        return response

    def __lookup_fs_session(self, url):
        # fast direct matches
        if url in self.__sessions:
            return self.__sessions[url]
        # slower pattern search depends on pattern count and size
        for k in self.__sessions_keys:
            if url.find(k) == 0:
                return self.__sessions[k]
        return None

    def request(self, *args, **kwargs):
        session = self.__lookup_fs_session(args[1])
        if not session is None:
            return session.request(*args, **kwargs)
        return super(FuturesSessionFlex, self).request(*args, **kwargs)

    def close(self):
        try:
            for key, session in self.__sessions.items():
                session.close()
        except Exception:
            LOGGER.exception("Error closing sessions")
